@@ -6,25 +6,38 @@ from __future__ import division, print_function
 import os
 from os.path import (join as pjoin, dirname, basename, exists, abspath,
                      relpath, realpath)
+import logging
 import shutil
 import warnings
 from subprocess import Popen, PIPE
+from typing import (Callable, Dict, FrozenSet, Iterable, List, Mapping,
+                    Optional, Set, Text, Tuple, Union)
 
 from .pycompat import string_types
-from .libsana import tree_libs, stripped_lib_dict, get_rp_stripper
-from .tools import (set_install_name, zip2dir, dir2zip,
+from .libsana import (tree_libs, stripped_lib_dict, get_rp_stripper,
+                      walk_directory, get_dependencies)
+from .tools import (set_install_name, zip2dir, dir2zip, validate_signature,
                     find_package_dirs, set_install_id, get_archs)
-from .tmpdirs import InTemporaryDirectory, InGivenDirectory
+from .tmpdirs import TemporaryDirectory
 from .wheeltools import rewrite_record, InWheel
 
+
+logger = logging.getLogger(__name__)
+
 # Prefix for install_name_id of copied libraries
 DLC_PREFIX = '/DLC/'
 
+
 class DelocationError(Exception):
     pass
 
 
-def delocate_tree_libs(lib_dict, lib_path, root_path):
+def delocate_tree_libs(
+    lib_dict,  # type: Mapping[Text, Mapping[Text, Text]]
+    lib_path,  # type: Text
+    root_path  # type: Text
+):
+    # type: (...) -> Dict[Text, Dict[Text, Text]]
     """ Move needed libraries in `lib_dict` into `lib_path`
 
     `lib_dict` has keys naming libraries required by the files in the
@@ -46,7 +59,7 @@ def delocate_tree_libs(lib_dict, lib_path, root_path):
     lib_path : str
         Path in which to store copies of libs referred to in keys of
         `lib_dict`.  Assumed to exist
-    root_path : str, optional
+    root_path : str
         Root directory of tree analyzed in `lib_dict`.  Any required
         library within the subtrees of `root_path` does not get copied, but
         libraries linking to it have links adjusted to use relative path to
@@ -57,19 +70,51 @@ def delocate_tree_libs(lib_dict, lib_path, root_path):
     copied_libs : dict
         Filtered `lib_dict` dict containing only the (key, value) pairs from
         `lib_dict` where the keys are the libraries copied to `lib_path``.
+
+    Raises
+    ------
+    DelocationError
+        When a malformed `lib_dict` has unresolved paths, missing files, etc.
+        When two dependencies would've been be copied to the same destination.
     """
-    copied_libs = {}
-    delocated_libs = set()
+    # Test for errors first to avoid getting half-way through changing the tree
+    libraries_to_copy, libraries_to_delocate = _analyze_tree_libs(
+        lib_dict, root_path
+    )
+    # Copy libraries and update lib_dict.
+    lib_dict, copied_libraries = _copy_required_libs(
+        lib_dict, lib_path, root_path, libraries_to_copy
+    )
+    # Update the install names of local and copied libaries.
+    _update_install_names(
+        lib_dict, root_path, libraries_to_delocate | copied_libraries
+    )
+    return libraries_to_copy
+
+
+def _analyze_tree_libs(
+    lib_dict,  # type: Mapping[Text, Mapping[Text, Text]]
+    root_path  # type: Text
+):
+    # type: (...) -> Tuple[Dict[Text, Dict[Text, Text]], Set[Text]]
+    """ Verify then return which library files to copy and delocate.
+
+    Returns
+    -------
+    needs_copying : dict
+        The libraries outside of `root_path`.
+        This is in the `lib_dict` format for use by `delocate_tree_libs`.
+    needs_delocating : set of str
+        The libraries inside of `root_path` which need to be delocated.
+    """
+    needs_delocating = set()  # Libraries which need install names updated.
+    needs_copying = {}  # A report of which libraries were copied.
     copied_basenames = set()
     rp_root_path = realpath(root_path)
-    rp_lib_path = realpath(lib_path)
-    # Test for errors first to avoid getting half-way through changing the tree
     for required, requirings in lib_dict.items():
-        if required.startswith('@'): # assume @rpath etc are correct
-            # But warn, because likely they are not
-            warnings.warn('Not processing required path {0} because it '
-                          'begins with @'.format(required))
-            continue
+        if required.startswith('@'):
+            # @rpath, etc, at this point should never happen.
+            raise DelocationError("%s was expected to be resolved." % required)
         r_ed_base = basename(required)
         if relpath(required, rp_root_path).startswith('..'):
             # Not local, plan to copy
@@ -79,33 +124,81 @@ def delocate_tree_libs(lib_dict, lib_path, root_path):
             if not exists(required):
                 raise DelocationError('library "{0}" does not exist'.format(
                     required))
-            copied_libs[required] = requirings
+            # Copy requirings to preserve it since it will be modified later.
+            needs_copying[required] = dict(requirings)
             copied_basenames.add(r_ed_base)
-        else: # Is local, plan to set relative loader_path
-            out_path = pjoin(lib_path, basename(required))
-            if exists(out_path):
-                raise DelocationError(
-                    'Output {0} already exists'.format(out_path))
-            delocated_libs.add(required)
-    # Modify in place now that we've checked for errors
-    for required in copied_libs:
-        shutil.copy2(required, lib_path)
-        # Set rpath and install names for this copied library
-        for requiring, orig_install_name in lib_dict[required].items():
-            req_rel = relpath(rp_lib_path, dirname(requiring))
-            set_install_name(requiring, orig_install_name,
-                             '@loader_path/{0}/{1}'.format(
-                                 req_rel, basename(required)))
-    for required in delocated_libs:
+        else:  # Is local, plan to set relative loader_path
+            needs_delocating.add(required)
+    return needs_copying, needs_delocating
+
+
+def _copy_required_libs(
+    lib_dict,  # type: Mapping[Text, Mapping[Text, Text]]
+    lib_path,  # type: Text
+    root_path,  # type: Text
+    libraries_to_copy  # type: Iterable[Text]
+):
+    # type (...) -> Tuple[Dict[Text, Dict[Text, Text]], Set[Text]]
+    """ Copy libraries outside of root_path to lib_path.
+
+    Returns
+    -------
+    updated_lib_dict : dict
+        A copy of `lib_dict` modified so that dependencies now point to
+        the copied library destinations.
+    needs_delocating : set of str
+        A set of the destination files, these need to be delocated.
+    """
+    # Create a copy of lib_dict for this script to modify and return.
+    out_lib_dict = _copy_lib_dict(lib_dict)
+    del lib_dict
+    needs_delocating = set()  # Set[Text]
+    for old_path in libraries_to_copy:
+        new_path = realpath(pjoin(lib_path, basename(old_path)))
+        logger.info(
+            "Copying library %s to %s", old_path, relpath(new_path, root_path)
+        )
+        shutil.copy(old_path, new_path)
+        # Delocate this file now that it is stored locally.
+        needs_delocating.add(new_path)
+        # Update out_lib_dict with the new file paths.
+        out_lib_dict[new_path] = out_lib_dict[old_path]
+        del out_lib_dict[old_path]
+        for required in out_lib_dict:
+            if old_path not in out_lib_dict[required]:
+                continue
+            out_lib_dict[required][new_path] = out_lib_dict[required][old_path]
+            del out_lib_dict[required][old_path]
+    return out_lib_dict, needs_delocating
+
+
+def _update_install_names(
+    lib_dict,  # type: Mapping[Text, Mapping[Text, Text]]
+    root_path,  # type: Text
+    files_to_delocate  # type: Iterable[Text]
+):
+    # type: (...) -> None
+    """ Update the install names of libraries. """
+    for required in files_to_delocate:
         # Set relative path for local library
         for requiring, orig_install_name in lib_dict[required].items():
             req_rel = relpath(required, dirname(requiring))
-            set_install_name(requiring, orig_install_name,
-                             '@loader_path/' + req_rel)
-    return copied_libs
+            new_install_name = '@loader_path/' + req_rel
+            logger.info(
+                "Modifying install name in %s from %s to %s",
+                relpath(requiring, root_path),
+                orig_install_name,
+                new_install_name,
+            )
+            set_install_name(requiring, orig_install_name, new_install_name)
 
 
-def copy_recurse(lib_path, copy_filt_func = None, copied_libs = None):
+def copy_recurse(
+    lib_path,  # type: Text
+    copy_filt_func=None,  # type: Optional[Callable[[Text], bool]]
+    copied_libs=None  # type: Optional[Dict[Text, Dict[Text, Text]]]
+):
+    # type: (...) -> Dict[Text, Dict[Text, Text]]
     """ Analyze `lib_path` for library dependencies and copy libraries
 
     `lib_path` is a directory containing libraries.  The libraries might
@@ -138,7 +231,16 @@ def copy_recurse(lib_path, copy_filt_func = None, copied_libs = None):
     copied_libs : dict
         Input `copied_libs` dict with any extra libraries and / or dependencies
         added.
+
+    .. deprecated:: 0.9
+        This function is obsolete.  :func:`delocate_path` handles recursive
+        dependencies while also supporting `@loader_path`.
     """
+    warnings.warn(
+        "copy_recurse is obsolete and should no longer be called.",
+        DeprecationWarning,
+        stacklevel=2,
+    )
     if copied_libs is None:
         copied_libs = {}
     else:
@@ -151,8 +253,13 @@ def copy_recurse(lib_path, copy_filt_func = None, copied_libs = None):
     return copied_libs
 
 
-def _copy_required(lib_path, copy_filt_func, copied_libs):
-    """ Copy libraries required for files in `lib_path` to `lib_path`
+def _copy_required(
+    lib_path,  # type: Text
+    copy_filt_func,  # type: Optional[Callable[[Text], bool]]
+    copied_libs  # type: Dict[Text, Dict[Text, Text]]
+):
+    # type: (...) -> None
+    """ Copy libraries required for files in `lib_path` to `copied_libs`
 
     Augment `copied_libs` dictionary with any newly copied libraries, modifying
     `copied_libs` in-place - see Notes.
@@ -200,6 +307,9 @@ def _copy_required(lib_path, copy_filt_func, copied_libs):
     ``/sys/libB.dylib``, so we fix our copy of `libC.dylib`` to point to
     ``my_lib_path/libB.dylib`` and add ``/sys/libC.dylib`` as a
     ``dependings_dict`` entry for ``copied_libs['/sys/libB.dylib']``
+
+    .. deprecated:: 0.9
+        This function is obsolete, and is only used by :func:`copy_recurse`.
     """
     # Paths will be prepended with `lib_path`
     lib_dict = tree_libs(lib_path)
@@ -207,7 +317,7 @@ def _copy_required(lib_path, copy_filt_func, copied_libs):
     rp_lp = realpath(lib_path)
     copied2orig = dict((pjoin(rp_lp, basename(c)), c) for c in copied_libs)
     for required, requirings in lib_dict.items():
-        if not copy_filt_func is None and not copy_filt_func(required):
+        if copy_filt_func is not None and not copy_filt_func(required):
             continue
         if required.startswith('@'):
             # May have been processed by us, or have some rpath, loader_path of
@@ -232,24 +342,32 @@ def _copy_required(lib_path, copy_filt_func, copied_libs):
         out_path = pjoin(lib_path, basename(required))
         if exists(out_path):
             raise DelocationError(out_path + ' already exists')
-        shutil.copy2(required, lib_path)
+        shutil.copy(required, lib_path)
         copied2orig[out_path] = required
         copied_libs[required] = procd_requirings
 
 
 def _dylibs_only(filename):
+    # type: (Text) -> bool
     return (filename.endswith('.so') or
             filename.endswith('.dylib'))
 
 
 def filter_system_libs(libname):
+    # type: (Text) -> bool
     return not (libname.startswith('/usr/lib') or
                 libname.startswith('/System'))
 
 
-def delocate_path(tree_path, lib_path,
-                  lib_filt_func = _dylibs_only,
-                  copy_filt_func = filter_system_libs):
+def delocate_path(
+    tree_path,  # type: Text
+    lib_path,  # type: Text
+    lib_filt_func=None,  # type: Optional[Union[str, Callable[[Text], bool]]]
+    copy_filt_func=filter_system_libs,  # type: Optional[Callable[[Text], bool]]
+    executable_path=None,  # type: Optional[Text]
+    ignore_missing=False,  # type: bool
+):
+    # type: (...) -> Dict[Text, Dict[Text, Text]]
     """ Copy required libraries for files in `tree_path` into `lib_path`
 
     Parameters
@@ -258,17 +376,22 @@ def delocate_path(tree_path, lib_path,
         Root path of tree to search for required libraries
     lib_path : str
         Directory into which we copy required libraries
-    lib_filt_func : None or callable, optional
+    lib_filt_func : None or str or callable, optional
         If None, inspect all files for dependencies on dynamic libraries. If
         callable, accepts filename as argument, returns True if we should
-        inspect the file, False otherwise. Default is callable rejecting all
-        but files ending in ``.so`` or ``.dylib``.
+        inspect the file, False otherwise. If str == "dylibs-only" then inspect
+        only files with known dynamic library extensions (``.dylib``, ``.so``).
     copy_filt_func : None or callable, optional
         If callable, called on each library name detected as a dependency; copy
         where ``copy_filt_func(libname)`` is True, don't copy otherwise.
         Default is callable rejecting only libraries beginning with
         ``/usr/lib`` or ``/System``.  None means copy all libraries. This will
         usually end up copying large parts of the system run-time.
+    executable_path : None or str, optional
+        If not None, an alternative path to use for resolving
+        `@executable_path`.
+    ignore_missing : bool, default=False
+        Continue even if missing dependencies are detected.
 
     Returns
     -------
@@ -280,36 +403,80 @@ def delocate_path(tree_path, lib_path,
         is a file in the path depending on ``copied_lib_path``, and the value
         is the ``install_name`` of ``copied_lib_path`` in the depending
         library.
+
+    Raises
+    ------
+    DelocationError
+        When any dependencies can not be located.
     """
+    if lib_filt_func == "dylibs-only":
+        lib_filt_func = _dylibs_only
+    elif isinstance(lib_filt_func, str):
+        raise TypeError('lib_filt_func string can only be "dylibs-only"')
+    if lib_filt_func is None:
+        lib_filt_func = (lambda _: True)
     if not exists(lib_path):
         os.makedirs(lib_path)
-    lib_dict = tree_libs(tree_path, lib_filt_func)
-    if not copy_filt_func is None:
-        lib_dict = dict((key, value) for key, value in lib_dict.items()
-                        if copy_filt_func(key))
-    copied = delocate_tree_libs(lib_dict, lib_path, tree_path)
-    return copy_recurse(lib_path, copy_filt_func, copied)
+
+    lib_dict = {}  # type: Dict[Text, Dict[Text, Text]]
+    missing_libs = False
+    for library_path in walk_directory(
+        tree_path, lib_filt_func, executable_path=executable_path
+    ):
+        for depending_path, install_name in get_dependencies(
+            library_path,
+            executable_path=executable_path,
+            filt_func=lib_filt_func,
+        ):
+            if depending_path is None:
+                missing_libs = True
+                continue
+            if copy_filt_func and not copy_filt_func(depending_path):
+                continue
+            lib_dict.setdefault(depending_path, {})
+            lib_dict[depending_path][library_path] = install_name
+
+    if missing_libs and not ignore_missing:
+        # Details of missing libraries would have already reported by
+        # get_dependencies.
+        raise DelocationError("Could not find all dependencies.")
+
+    return delocate_tree_libs(lib_dict, lib_path, tree_path)
+
+
+def _copy_lib_dict(lib_dict):
+    # type: (Mapping[Text, Mapping[Text, Text]]) -> Dict[Text, Dict[Text, Text]]  # noqa: E501
+    """ Returns a copy of lib_dict. """
+    return {  # Convert nested Mapping types into nested Dict types.
+        required: dict(requiring) for required, requiring in lib_dict.items()
+    }
 
 
 def _merge_lib_dict(d1, d2):
+    # type: (Dict[Text, Dict[Text, Text]], Mapping[Text, Mapping[Text, Text]]) -> None  # noqa: E501
     """ Merges lib_dict `d2` into lib_dict `d1`
     """
     for required, requirings in d2.items():
         if required in d1:
             d1[required].update(requirings)
         else:
-            d1[required] = requirings
+            # Convert to dict to satisfy type checking.
+            d1[required] = dict(requirings)
     return None
 
 
-def delocate_wheel(in_wheel,
-                   out_wheel = None,
-                   lib_sdir = '.dylibs',
-                   lib_filt_func = _dylibs_only,
-                   copy_filt_func = filter_system_libs,
-                   require_archs = None,
-                   check_verbose = False,
-                  ):
+def delocate_wheel(
+    in_wheel,  # type: Text
+    out_wheel=None,  # type: Optional[Text]
+    lib_sdir='.dylibs',  # type: Text
+    lib_filt_func=None,  # type: Union[None, str, Callable[[Text], bool]]
+    copy_filt_func=filter_system_libs,  # type: Optional[Callable[[Text], bool]]
+    require_archs=None,  # type: Union[None, Text, Iterable[Text]]
+    check_verbose=False,  # type: bool
+    executable_path=None,  # type: Optional[Text]
+    ignore_missing=False,  # type: bool
+):
+    # type: (...) -> Dict[Text, Dict[Text, Text]]
     """ Update wheel by copying required libraries to `lib_sdir` in wheel
 
     Create `lib_sdir` in wheel tree only if we are copying one or more
@@ -327,11 +494,11 @@ def delocate_wheel(in_wheel,
     lib_sdir : str, optional
         Subdirectory name in wheel package directory (or directories) to store
         needed libraries.
-    lib_filt_func : None or callable, optional
+    lib_filt_func : None or str or callable, optional
         If None, inspect all files for dependencies on dynamic libraries. If
         callable, accepts filename as argument, returns True if we should
-        inspect the file, False otherwise. Default is callable rejecting all
-        but files ending in ``.so`` or ``.dylib``.
+        inspect the file, False otherwise. If str == "dylibs-only" then inspect
+        only files with known dynamic library extensions (``.dylib``, ``.so``).
     copy_filt_func : None or callable, optional
         If callable, called on each library name detected as a dependency; copy
         where ``copy_filt_func(libname)`` is True, don't copy otherwise.
@@ -348,6 +515,10 @@ def delocate_wheel(in_wheel,
         (e.g "i386" or "x86_64").
     check_verbose : bool, optional
         If True, print warning messages about missing required architectures
+    executable_path : str, optional
+        An alternative path to use for resolving `@executable_path`.
+    ignore_missing : bool, default=False
+        Continue even if missing dependencies are detected.
 
     Returns
     -------
@@ -366,45 +537,55 @@ def delocate_wheel(in_wheel,
     else:
         out_wheel = abspath(out_wheel)
     in_place = in_wheel == out_wheel
-    with InTemporaryDirectory() as tmpdir:
-        all_copied = {}
-        zip2dir(in_wheel, 'wheel')
-        with InGivenDirectory('wheel'):
-            for package_path in find_package_dirs('.'):
-                lib_path = pjoin(package_path, lib_sdir)
-                lib_path_exists = exists(lib_path)
-                copied_libs = delocate_path(package_path, lib_path,
-                                            lib_filt_func, copy_filt_func)
-                if copied_libs and lib_path_exists:
+    with TemporaryDirectory() as tmpdir:
+        all_copied = {}  # type: Dict[Text, Dict[Text, Text]]
+        wheel_dir = realpath(pjoin(tmpdir, 'wheel'))
+        zip2dir(in_wheel, wheel_dir)
+        for package_path in find_package_dirs(wheel_dir):
+            lib_path = pjoin(package_path, lib_sdir)
+            lib_path_exists = exists(lib_path)
+            copied_libs = delocate_path(
+                package_path,
+                lib_path,
+                lib_filt_func,
+                copy_filt_func,
+                executable_path=executable_path,
+                ignore_missing=ignore_missing,
+            )
+            if copied_libs and lib_path_exists:
+                raise DelocationError(
+                    '{0} already exists in wheel but need to copy '
+                    '{1}'.format(lib_path, '; '.join(copied_libs)))
+            if len(os.listdir(lib_path)) == 0:
+                shutil.rmtree(lib_path)
+            # Check architectures
+            if require_archs is not None:
+                stop_fast = not check_verbose
+                bads = check_archs(copied_libs, require_archs, stop_fast)
+                if len(bads) != 0:
+                    if check_verbose:
+                        print(bads_report(bads, pjoin(tmpdir, 'wheel')))
                     raise DelocationError(
-                        '{0} already exists in wheel'.format(lib_path))
-                if len(os.listdir(lib_path)) == 0:
-                    shutil.rmtree(lib_path)
-                # Check architectures
-                if not require_archs is None:
-                    stop_fast = not check_verbose
-                    bads = check_archs(copied_libs, require_archs, stop_fast)
-                    if len(bads) != 0:
-                        if check_verbose:
-                            print(bads_report(bads, pjoin(tmpdir, 'wheel')))
-                        raise DelocationError(
-                            "Some missing architectures in wheel")
-                # Change install ids to be unique within Python space
-                install_id_root = DLC_PREFIX + package_path + '/'
-                for lib in copied_libs:
-                    lib_base = basename(lib)
-                    copied_path = pjoin(lib_path, lib_base)
-                    set_install_id(copied_path, install_id_root + lib_base)
-                _merge_lib_dict(all_copied, copied_libs)
+                        "Some missing architectures in wheel")
+            # Change install ids to be unique within Python space
+            install_id_root = (DLC_PREFIX +
+                               relpath(package_path, wheel_dir) +
+                               '/')
+            for lib in copied_libs:
+                lib_base = basename(lib)
+                copied_path = pjoin(lib_path, lib_base)
+                set_install_id(copied_path, install_id_root + lib_base)
+                validate_signature(copied_path)
+            _merge_lib_dict(all_copied, copied_libs)
         if len(all_copied):
-            rewrite_record('wheel')
+            rewrite_record(wheel_dir)
         if len(all_copied) or not in_place:
-            dir2zip('wheel', out_wheel)
-    wheel_dir = realpath(pjoin(tmpdir, 'wheel'))
+            dir2zip(wheel_dir, out_wheel)
     return stripped_lib_dict(all_copied, wheel_dir + os.path.sep)
 
 
 def patch_wheel(in_wheel, patch_fname, out_wheel=None):
+    # type: (Text, Text, Optional[Text]) -> None
     """ Apply ``-p1`` style patch in `patch_fname` to contents of `in_wheel`
 
     If `out_wheel` is None (the default), overwrite the wheel `in_wheel`
@@ -431,16 +612,27 @@ def patch_wheel(in_wheel, patch_fname, out_wheel=None):
     with InWheel(in_wheel, out_wheel):
         with open(patch_fname, 'rb') as fobj:
             patch_proc = Popen(['patch', '-p1'],
-                               stdin = fobj,
-                               stdout = PIPE,
-                               stderr = PIPE)
+                               stdin=fobj,
+                               stdout=PIPE,
+                               stderr=PIPE)
             stdout, stderr = patch_proc.communicate()
             if patch_proc.returncode != 0:
                 raise RuntimeError("Patch failed with stdout:\n" +
                                    stdout.decode('latin1'))
 
 
-def check_archs(copied_libs, require_archs=(), stop_fast=False):
+_ARCH_LOOKUP = {
+    'intel': ['i386', 'x86_64'],
+    'universal2': ['x86_64', 'arm64']
+}
+
+
+def check_archs(
+    copied_libs,  # type: Mapping[Text, Mapping[Text, Text]]
+    require_archs=(),  # type:  Union[Text, Iterable[Text]]
+    stop_fast=False  # type: bool
+):
+    # type: (...) -> Set[Union[Tuple[Text, FrozenSet[Text]], Tuple[Text, Text, FrozenSet[Text]]]]  # noqa: E501
     """ Check compatibility of archs in `copied_libs` dict
 
     Parameters
@@ -460,8 +652,9 @@ def check_archs(copied_libs, require_archs=(), stop_fast=False):
         these architectures are. If a sequence, then a set of required
         architectures e.g. ``['i386', 'x86_64']`` to specify dual Intel
         architectures.  If a string, then a standard architecture name as
-        returned by ``lipo -info`` or the string "intel", corresponding to the
-        sequence ``['i386', 'x86_64']``
+        returned by ``lipo -info``, or the string "intel", corresponding to the
+        sequence ``['i386', 'x86_64']``, or the string "universal2",
+        corresponding to ``['x86_64', 'arm64']``.
     stop_fast : bool, optional
         Whether to give up collecting errors after the first
 
@@ -480,19 +673,18 @@ def check_archs(copied_libs, require_archs=(), stop_fast=False):
         required.
     """
     if isinstance(require_archs, string_types):
-        require_archs = (['i386', 'x86_64'] if require_archs == 'intel'
-                         else [require_archs])
-    require_archs = frozenset(require_archs)
-    bads = []
+        require_archs = _ARCH_LOOKUP.get(require_archs, [require_archs])
+    require_archs_set = frozenset(require_archs)
+    bads = []  # type: List[Union[Tuple[Text, FrozenSet[Text]], Tuple[Text, Text, FrozenSet[Text]]]]  # noqa: E501
     for depended_lib, dep_dict in copied_libs.items():
         depended_archs = get_archs(depended_lib)
         for depending_lib, install_name in dep_dict.items():
             depending_archs = get_archs(depending_lib)
-            all_required = depending_archs | require_archs
+            all_required = depending_archs | require_archs_set
             all_missing = all_required.difference(depended_archs)
             if len(all_missing) == 0:
                 continue
-            required_missing = require_archs.difference(depended_archs)
+            required_missing = require_archs_set.difference(depended_archs)
             if len(required_missing):
                 bads.append((depending_lib, required_missing))
             else:
@@ -527,7 +719,7 @@ def bads_report(bads, path_prefix=None):
     report : str
         A nice report for printing
     """
-    path_processor = ((lambda x : x) if path_prefix is None
+    path_processor = ((lambda x: x) if path_prefix is None
                       else get_rp_stripper(path_prefix))
     reports = []
     for result in bads:
