@@ -2,7 +2,7 @@
 # Create libs used for testing
 # Run in directory containing this file
 # With thanks to https://dev.lsstcorp.org/trac/wiki/LinkingDarwin
-# I ran this on a Snow Leopard machine with CXX=g++ ./make_libs.sh
+export MACOSX_DEPLOYMENT_TARGET=10.9
 
 if [ "$CXX" = "" ]; then
     CXX=c++
@@ -31,7 +31,7 @@ int main(int, char**) { c(); return 0; }
 EOF
 
 CXX_64="$CXX -arch x86_64"
-CXX_32="$CXX -arch i386"
+CXX_M1="$CXX -arch arm64"
 
 # Delete previous before rebuilding
 if [ -e liba.dylib ]; then
@@ -44,7 +44,7 @@ if [ -e libb.dylib ]; then
 fi
 
 $CXX_64 -o liba.dylib -dynamiclib a.cc
-$CXX_32 -o liba32.dylib -dynamiclib a.cc
+$CXX_M1 -o libam1.dylib -dynamiclib a.cc
 $CXX_64 -o a.o -c a.cc
 ar rcs liba.a a.o
 $CXX_64 -o libb.dylib -dynamiclib b.cc -L. -la
@@ -52,6 +52,15 @@ $CXX_64 -o libb.dylib -dynamiclib b.cc -L. -la
 $CXX_64 -o libc.dylib -dynamiclib c.cc -L. -la -lb
 $CXX_64 -o test-lib d.cc -L. -lc
 
+# Make a dual-arch library
+lipo -create liba.dylib libam1.dylib -output liba_both.dylib
+
+# Make a library with architecture name in otool -L line
+lipo -create libam1.dylib -output libam1-arch.dylib
+
 # Change permissions in nasty way to test working with permissions
 chmod 444 liba.dylib
 chmod 400 libb.dylib
+
+# Remove temporary files
+rm *.cc
